#!/usr/bin/python3
import logging
import random
import os
import time
from collections import Counter
from collections import defaultdict

import numpy as np
import pandas as pd
from scipy.signal import find_peaks
import matplotlib
import matplotlib.pyplot as plt
from gamutrf.utils import SCAN_FRES, SCAN_FROLL, WIDTH, HEIGHT, DPI, MPL_BACKEND


ROLLOVERHZ = 100e6
CSV = ".csv"
CSVCOLS = ["ts", "freq", "db"]
ROLLING_FACTOR = int(SCAN_FROLL / SCAN_FRES)

logging.getLogger("matplotlib.font_manager").disabled = True


def read_csv_chunks(args):
    minmhz = args.minhz / 1e6
    leftover_df = pd.DataFrame()

    def detect_frames(df):
        freqdiff = df["freq"].diff().abs()
        df["frame"] = 0
        # Detect tuning wraparound, where frequency changed by more than 100MHz
        df.loc[freqdiff > (ROLLOVERHZ / 1e6), ["frame"]] = 1
        df[
            "frame"
        ] = (  # pylint: disable=unsupported-assignment-operation,disable=unsubscriptable-object
            df["frame"].cumsum().fillna(0).astype(np.uint64)
        )  # pylint: disable=unsubscriptable-object

    def preprocess_frames(df):
        df.set_index("frame", inplace=True)
        df["ts"] = df.groupby("frame", sort=False)["ts"].transform(min)

    with pd.read_csv(
        args.csv, header=None, delim_whitespace=True, chunksize=args.nrows
    ) as reader:
        for chunk in reader:
            read_rows = len(chunk)
            logging.info(f"read chunk of {read_rows} from {args.csv}")
            chunk.columns = CSVCOLS
            chunk["freq"] /= 1e6
            df = pd.concat([leftover_df, chunk])
            detect_frames(df)
            read_frames = df["frame"].max()  # pylint: disable=unsubscriptable-object
            if read_frames == 0 and len(chunk) == args.nrows:
                leftover_df = leftover_df.append(chunk[CSVCOLS])
                logging.info(f"buffering incomplete frame - {args.nrows} too small?")
                continue
            leftover_df = df[df["frame"] == read_frames][
                CSVCOLS
            ]  # pylint: disable=unsubscriptable-object
            df = df[
                (df["frame"] < read_frames) & (df["freq"] >= minmhz)
            ]  # pylint: disable=unsubscriptable-object
            df = calc_db(df, args.db_rolling_factor)
            df = df[df["db"] >= args.mindb]
            preprocess_frames(df)
            yield df

    if len(leftover_df):
        df = leftover_df
        detect_frames(df)
        df = df[
            (df["frame"] < read_frames) & (df["freq"] >= minmhz)
        ]  # pylint: disable=unsubscriptable-object
        df = calc_db(df, args.db_rolling_factor)
        df = df[df["db"] >= args.mindb]
        preprocess_frames(df)
        yield df


def read_csv(args):
    frames = 0

    for df in read_csv_chunks(args):
        for _, frame_df in df.groupby("frame"):
            yield (frames, frame_df)
            if args.maxframe and frames == args.maxframe:
                return
            frames += 1


def choose_recorders(signals, recorder_freq_exclusions, max_recorder_signals):
    suitable_recorders = defaultdict(set)
    for signal in sorted(signals):
        for recorder, excluded in sorted(recorder_freq_exclusions.items()):
            if not freq_excluded(signal, excluded):
                suitable_recorders[signal].add(recorder)
    recorder_assignments = []
    busy_count = defaultdict(int)
    for signal, recorders in sorted(suitable_recorders.items(), key=lambda x: x[1]):
        if not recorders:
            continue
        busy_recorders = set(
            recorder
            for recorder, count in busy_count.items()
            if count >= max_recorder_signals
        )
        free_recorders = recorders - busy_recorders
        if not free_recorders:
            continue
        recorder = random.choice(list(free_recorders))  # nosec
        busy_count[recorder] += 1
        recorder_assignments.append((signal, recorder))
    return recorder_assignments


def parse_freq_excluded(freq_exclusions_raw):
    freq_exclusions = []
    for pair in freq_exclusions_raw:
        freq_min, freq_max = pair.split("-")
        if len(freq_min):
            freq_min = int(freq_min)
        else:
            freq_min = None
        if len(freq_max):
            freq_max = int(freq_max)
        else:
            freq_max = None
        freq_exclusions.append((freq_min, freq_max))
    return tuple(freq_exclusions)


def freq_excluded(freq, freq_exclusions):
    for freq_min, freq_max in freq_exclusions:
        if freq_min is not None and freq_max is not None:
            if freq >= freq_min and freq <= freq_max:
                return True
            continue
        if freq_min is None:
            if freq <= freq_max:
                return True
            continue
        if freq >= freq_min:
            return True
    return False


def calc_db(df, rolling_factor=ROLLING_FACTOR):
    df.replace([np.inf, -np.inf], np.nan, inplace=True)
    meandb = df["db"].mean()
    if rolling_factor:
        df["db"] = df["db"].rolling(rolling_factor).mean().fillna(meandb)
    return df


def scipy_find_sig_windows(df, width, prominence, threshold):
    data = df.db.to_numpy()
    peaks, _ = find_peaks(data, prominence=prominence, width=(width,), height=threshold)
    return [(df.iloc[peak].freq, df.iloc[peak].db) for peak in peaks]


def graph_fft_peaks(
    graph_path, df, mean_running_df, sample_count_df, signals, last_dfs, scan_config
):
    maxdb = df.db.max()
    df["peaks"] = df.db.min()
    for peak_freq, _ in signals:
        df.loc[df.freq == peak_freq, "peaks"] = maxdb

    peak_df = df[df.peaks == maxdb].sort_values("db", ascending=False)[:5]
    peak_signals = ",".join(
        ["%.1f MHz %.1f dB" % (row.freq, row.db) for row in peak_df.itertuples()]
    )
    if peak_signals:
        peak_signals = f"strongest peak signals: {peak_signals}"

    matplotlib.use(MPL_BACKEND)
    plt.figure(figsize=(WIDTH, HEIGHT), dpi=DPI)
    plt.plot(
        df.freq,
        df.db,
        "b",
        df.freq,
        df.peaks,
        "y",
        mean_running_df.freq,
        mean_running_df.db,
        "k",
        # sample_count_df["freq"],
        # sample_count_df["size"],
        # "c",
    )
    for x, y in last_dfs:
        plt.plot(x, y)
    plt.xlabel("freq (MHz)")
    plt.ylabel("power (dB)")
    plt.legend(("power", "peak status", "mean power"), loc="upper right")
    ts_min = df.ts.min()
    ts_max = df.ts.max()
    time_min = time.ctime(ts_min)
    time_max = time.ctime(ts_max)
    duration = ts_max - ts_min
    scan_config_txt = ", ".join([f"{x}: {y}" for x, y in scan_config.items()])
    title = f"gamutRF scanner FFT {time_min} to {time_max}, {duration}s\n{scan_config_txt}\n{peak_signals}"
    plt.title(title)
    real_path = os.path.realpath(graph_path)
    basename = os.path.basename(real_path)
    dirname = os.path.dirname(real_path)
    tmp_graph_path = os.path.join(dirname, "." + basename)
    plt.savefig(tmp_graph_path)
    plt.cla()
    plt.close("all")
    os.rename(tmp_graph_path, graph_path)


def get_center(signal_mhz, freq_start_mhz, bin_mhz, record_bw):
    return int(
        int((signal_mhz - freq_start_mhz) / record_bw) * bin_mhz + freq_start_mhz
    )


def choose_record_signal(signals, max_signals):
    recorder_buckets = Counter()

    # Convert signals into buckets of record_bw size, count how many of each size
    for bucket in signals:
        recorder_buckets[bucket] += 1

    # Now count number of buckets of each count.
    buckets_by_count = defaultdict(set)
    for bucket, count in recorder_buckets.items():
        buckets_by_count[count].add(bucket)

    recorder_freqs = []
    # From least occuring bucket to most occurring, choose a random bucket for each recorder.
    for _count, buckets in sorted(buckets_by_count.items()):
        while buckets and len(recorder_freqs) < max_signals:
            bucket = random.choice(list(buckets))  # nosec
            buckets = buckets.remove(bucket)
            recorder_freqs.append(bucket)
        if len(recorder_freqs) == max_signals:
            break
    return recorder_freqs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import sys

try:
    from gnuradio import blocks  # pytype: disable=import-error
    from gnuradio import fft  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio import zeromq  # pytype: disable=import-error
    from gnuradio.fft import window  # pytype: disable=import-error
except ModuleNotFoundError:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme)"
    )
    sys.exit(1)

from gamutrf.grsource import get_source


class grscan(gr.top_block):
    def __init__(
        self,
        freq_end=1e9,
        freq_start=100e6,
        igain=0,
        samp_rate=4.096e6,
        sweep_sec=30,
        logaddr="0.0.0.0",  # nosec
        logport=8001,
        sdr="ettus",
        sdrargs=None,
        fft_size=1024,
        tune_overlap=0.5,
        tune_step_fft=0,
        skip_tune_step_fft=0,
        fft_roll=False,
        iqtlabs=None,
    ):
        gr.top_block.__init__(self, "scan", catch_exceptions=True)

        ##################################################
        # Parameters
        ##################################################
        self.freq_end = freq_end
        self.freq_start = freq_start
        self.sweep_sec = sweep_sec
        self.fft_size = fft_size

        ##################################################
        # Blocks
        ##################################################

        logging.info(f"will scan from {freq_start} to {freq_end}")
        get_source(
            self,
            sdr,
            samp_rate,
            igain,
            agc=False,
            center_freq=freq_start,
            sdrargs=sdrargs,
        )

        self.retune_fft = None
        if iqtlabs:
            freq_range = freq_end - freq_start
            tune_step_hz = int(samp_rate * tune_overlap)
            if tune_step_fft:
                logging.info(
                    f"retuning across {freq_range/1e6}MHz every {tune_step_fft} FFTs"
                )
            else:
                target_retune_hz = freq_range / self.sweep_sec / tune_step_hz
                fft_rate = int(samp_rate / fft_size)
                tune_step_fft = int(fft_rate / target_retune_hz)
                logging.info(
                    f"retuning across {freq_range/1e6}MHz in {self.sweep_sec}s, requires retuning at {target_retune_hz}Hz in {tune_step_hz/1e6}MHz steps ({tune_step_fft} FFTs)"
                )
            self.retune_fft = iqtlabs.retune_fft(
                "rx_freq",
                fft_size,
                fft_size,
                int(samp_rate),
                int(freq_start),
                int(freq_end),
                tune_step_hz,
                tune_step_fft,
                skip_tune_step_fft,
                fft_roll,
                -100,
                50,
            )
        self.fft_vxx_0 = fft.fft_vcc(
            fft_size, True, window.blackmanharris(fft_size), True, 1
        )
        self.blocks_stream_to_vector_0 = blocks.stream_to_vector(
            gr.sizeof_gr_complex, fft_size
        )
        self.blocks_nlog10_ff_0 = blocks.nlog10_ff(20, fft_size, 0)
        zmq_addr = f"tcp://{logaddr}:{logport}"
        logging.info("serving FFT on %s", zmq_addr)
        self.zeromq_pub_sink_0 = zeromq.pub_sink(1, 1, zmq_addr, 100, False, 65536, "")
        self.blocks_complex_to_mag_0 = blocks.complex_to_mag(fft_size)

        ##################################################
        # Connections
        ##################################################
        # FFT chain
        if self.retune_fft:
            self.msg_connect((self.retune_fft, "tune"), (self.source_0, self.cmd_port))
            self.connect((self.blocks_stream_to_vector_0, 0), (self.fft_vxx_0, 0))
            self.connect((self.fft_vxx_0, 0), (self.blocks_complex_to_mag_0, 0))
            self.connect(
                (self.blocks_complex_to_mag_0, 0), (self.blocks_nlog10_ff_0, 0)
            )
            self.connect((self.blocks_nlog10_ff_0, 0), (self.retune_fft, 0))
            self.connect((self.retune_fft, 0), (self.zeromq_pub_sink_0, 0))
            self.connect((self.source_0, 0), (self.blocks_stream_to_vector_0, 0))



import argparse
import concurrent.futures
import logging
import os
import subprocess
import tempfile
import time

from gamutrf.sigwindows import CSV
from gamutrf.sigwindows import read_csv

MAX_WORKERS = 4


def generate_frames(args, tempdir):
    tsmap = []
    db_min = None
    db_max = None
    freq_min = None
    freq_max = None
    lastts = None

    def write_frame(frame, frame_f, frame_df):
        frame_fs = [frame_f]
        if frame in args.dumpframes:
            dumpfile = args.csv.replace(CSV, f"-{frame}.csv")
            frame_fs.append(dumpfile)
        for f in frame_fs:
            frame_df.to_csv(f, sep="\t", columns=["freq", "db"], index=False)

    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        maxmhz = args.maxhz / 1e6
        for frame, frame_df in read_csv(args):
            ts = frame_df["ts"].iat[0]
            frame_df = frame_df[frame_df["freq"] <= maxmhz]
            tsc = time.ctime(ts)
            frame_f = os.path.join(str(tempdir), str(ts))
            tsmap.append((frame, ts, tsc, frame_f))
            db = frame_df.db
            freq = frame_df.freq
            if frame:
                db_min = min(db_min, db.min())
                db_max = max(db_max, db.max())
                freq_min = min(freq_min, freq.min())
                freq_max = max(freq_max, freq.max())
            else:
                db_min = db.min()
                db_max = db.max()
                freq_min = freq.min()
                freq_max = freq.max()
                lastts = ts
            offset = ts - lastts
            logging.info(f"frame {frame} offset {offset}s at {tsc}")
            lastts = ts
            executor.submit(write_frame, frame, frame_f, frame_df.copy())
            frame_df.to_csv(
                os.path.join(str(tempdir), "graph.csv"),
                sep="\t",
                mode="a",
                columns=["ts", "freq", "db"],
                index=False,
                header=False,
            )
        executor.shutdown(wait=True)

    return (tsmap, freq_min, freq_max, db_min, db_max)


def call_gnuplot(gnuplot_cmds, tempdir):
    plot_f = os.path.join(str(tempdir), "plot.cmd")
    with open(plot_f, "w", encoding="utf8") as plot_c:
        plot_c.write("\n".join(gnuplot_cmds))

    logging.info("running gnuplot")
    subprocess.check_call(["gnuplot", plot_f])


def run_gnuplot(tsmap, freq_min, freq_max, db_min, db_max, args, tempdir, graph):
    logging.info("creating gunplot commands")
    y_min = db_min
    y_max = db_max
    freq_min = int(freq_min / 100) * 100
    freq_max = round(freq_max / 100) * 100
    xtics = (freq_max - freq_min) / args.xtics

    common_gnuplot_cmds = [
        "set terminal png truecolor rounded size 1920,720 enhanced",
        "set grid xtics",
        "set grid mxtics",
        "set grid ytics",
        "set xtics rotate by 90 right",
    ]

    gnuplot_cmds = common_gnuplot_cmds + [
        'set ytics format "%.4f"',
        'set xlabel "freq (MHz)"',
        'set ylabel "power (dB)"',
    ]

    gnuplot_cmds.extend(
        [
            f"set xtics {xtics}",
            f"set xrange [{freq_min}:{freq_max}]",
            f"set yrange [{y_min}:{y_max}]",
        ]
    )
    for frame, _ts, tsc, frame_f in tsmap:
        db_plot = f'"{frame_f}" using 1:2 with linespoints title "dB"'
        gnuplot_cmds.extend(
            [
                f'set output "{tempdir}/{frame:09}.png"',
                f'set title "{tsc} frame {frame}"',
                f"plot {db_plot}",
            ]
        )

    call_gnuplot(gnuplot_cmds, tempdir)

    gnuplot_cmds = common_gnuplot_cmds + [
        "set xdata time",
        'set timefmt "%s"',
        'set format x "%H:%M %d%m%y"',
        "set view 0,270",
        'set xlabel "time"',
        'set ylabel "freq (MHz)"',
        "set palette negative",
        "set pm3d map",
        f'set output "{graph}"',
        f'splot "{tempdir}/graph.csv" using 1:2:($3>{args.mindb}?$3:(1/0)) palette with circles title "dB"',
    ]

    call_gnuplot(gnuplot_cmds, tempdir)


def run_ffmpeg(args, tempdir, mp4):
    logging.info("running ffmpeg")
    subprocess.check_call(
        [
            "ffmpeg",
            "-loglevel",
            "error",
            "-framerate",
            str(args.framerate),
            "-pattern_type",
            "glob",
            "-i",
            os.path.join(str(tempdir), "*.png"),
            "-c:v",
            "libx264",
            "-pix_fmt",
            "yuv420p",
            "-r",
            "30",
            "-y",
            mp4,
        ]
    )


def main():
    parser = argparse.ArgumentParser(
        description="Convert a scan log to a timelapse graph/video"
    )
    parser.add_argument("csv", help="log file to parse")
    parser.add_argument(
        "--minhz", default=int(70 * 1e6), type=int, help="minimum frequency to process"
    )
    parser.add_argument(
        "--maxhz", default=int(6 * 1e9), type=int, help="maximum frequency to process"
    )
    parser.add_argument(
        "--mindb", default=int(-40), type=int, help="minimum dB to process"
    )
    parser.add_argument("--framerate", default=int(5), type=int, help="frame rate")
    parser.add_argument("--xtics", default=int(40), type=int, help="xtics")
    parser.add_argument(
        "--nrows", default=int(1e7), type=int, help="number of rows to read at once"
    )
    parser.add_argument("--tmproot", default="", help="root of temporary directory")
    parser.add_argument(
        "--maxframe", default=int(0), type=int, help="maximum frame number to process"
    )
    parser.add_argument(
        "--dumpframes",
        default=[],
        nargs="+",
        type=int,
        help="list of frame numbers to dump as csvs",
    )
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")

    if not args.csv.endswith(CSV):
        logging.fatal(f"{args.csv} must be {CSV}")
    if not os.path.exists(args.csv):
        logging.fatal(f"{args.csv} must exist")
    mp4 = args.csv.replace(CSV, ".mp4")
    graph = args.csv.replace(CSV, ".png")

    tmproot = None
    if args.tmproot:
        tmproot = args.tmproot

    with tempfile.TemporaryDirectory(dir=tmproot) as tempdir:
        tsmap, freq_min, freq_max, db_min, db_max = generate_frames(args, tempdir)
        run_gnuplot(tsmap, freq_min, freq_max, db_min, db_max, args, tempdir, graph)
        run_ffmpeg(args, tempdir, mp4)



import argparse
import datetime
import logging
import os
import time
from statistics import mean

import numpy as np

from gamutrf.sigwindows import CSV, ROLLING_FACTOR
from gamutrf.sigwindows import read_csv

MAX_WORKERS = 4


def generate_csv(args):
    with open(args.outcsv, "w", encoding="utf-8") as outcsv:
        for frame, frame_df in read_csv(args):
            ts = frame_df["ts"].iat[0]
            dt = datetime.datetime.fromtimestamp(ts)
            frame_df.freq *= 1e6
            frame_df.freq = frame_df.freq.astype(np.int64)
            freq = frame_df.freq
            logging.info(f"frame {frame} at {time.ctime(ts)}")
            freqmin = freq.min()
            freqmax = freq.max()
            steps = min(len(frame_df), args.fftmax)
            stephz = int((freqmax - freqmin) / steps)
            stepi = int(len(frame_df) / steps)
            line = [
                dt.strftime("%Y-%m-%d %H:%M:%S"),
                "",
                freqmin,
                freqmax,
                stephz,
                1,
            ]
            dbt = []
            for i, row in enumerate(frame_df.itertuples()):
                dbt.append(row.db)
                if i % stepi == 0:
                    line.append(mean(dbt))
                    dbt = []
            if dbt:
                line.append(mean(dbt))
            outcsv.write(",".join([str(i) for i in line]) + "\n")


def main():
    parser = argparse.ArgumentParser(
        description="Convert a scan log to a CSV file compatible with rtl-gopow (https://github.com/dhogborg/rtl-gopow)"
    )
    parser.add_argument("csv", help="log file to parse")
    parser.add_argument("outcsv", help="rtl_power style CSV output")
    parser.add_argument(
        "--minhz", default=int(70 * 1e6), type=int, help="minimum frequency to process"
    )
    parser.add_argument(
        "--maxhz", default=int(6 * 1e9), type=int, help="maximum frequency to process"
    )
    parser.add_argument(
        "--mindb", default=int(-40), type=int, help="minimum dB to process"
    )
    parser.add_argument(
        "--maxframe", default=int(0), type=int, help="maximum frame number to process"
    )
    parser.add_argument(
        "--nrows", default=int(1e7), type=int, help="number of rows to read at once"
    )
    parser.add_argument(
        "--fftmax", default=int(10000), type=int, help="max number of FFT bins"
    )
    parser.add_argument(
        "--db_rolling_factor",
        dest="db_rolling_factor",
        type=float,
        default=ROLLING_FACTOR,
        help="Divisor for rolling dB average (or 0 to disable)",
    )
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")

    if not args.csv.endswith(CSV):
        logging.fatal(f"{args.csv} must be {CSV}")
    if not os.path.exists(args.csv):
        logging.fatal(f"{args.csv} must exist")
    generate_csv(args)



import logging
import os
import sys
import time
from urllib.parse import urlparse

try:
    from gnuradio import blocks
    from gnuradio import soapy
    from gnuradio import uhd
    from gnuradio.gr import sizeof_gr_complex
except ModuleNotFoundError:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme)"
    )
    sys.exit(1)

from gamutrf.utils import ETTUS_ANT
from gamutrf.utils import ETTUS_ARGS


def set_command_time(_x, _y):
    raise NotImplementedError


def clear_command_time(_x):
    raise NotImplementedError


def get_sdr_time_now(_x):
    return time.time()


def get_source(
    grblock, sdr, samp_rate, gain, agc=False, center_freq=None, sdrargs=None
):
    logging.info(
        f"initializing SDR {sdr} with sample rate {samp_rate}, gain {gain}, agc {agc}"
    )

    grblock.set_command_time = set_command_time
    grblock.clear_command_time = clear_command_time
    grblock.get_sdr_time_now = get_sdr_time_now

    url = urlparse(sdr)
    if url.scheme:
        if url.scheme == "file" and os.path.exists(url.path):
            grblock.recording_source_0 = blocks.file_source(
                sizeof_gr_complex, url.path, True, 0, 0
            )
            grblock.source_0 = blocks.throttle(sizeof_gr_complex, samp_rate, True)
            grblock.connect((grblock.recording_source_0, 0), (grblock.source_0, 0))
            # TODO: enable setting frequency change tags on the stream, so can test scanner.
            grblock.freq_setter = lambda _x, _y: None
        else:
            raise ValueError("unsupported/missing file location")
        return

    if sdr == "ettus":
        if not sdrargs:
            sdrargs = ETTUS_ARGS
        grblock.source_0 = uhd.usrp_source(
            ",".join((sdrargs, "")),
            uhd.stream_args(
                cpu_format="fc32",
                args="",
                channels=list(range(0, 1)),
            ),
        )
        grblock.source_0.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        grblock.source_0.set_antenna(ETTUS_ANT, 0)
        grblock.source_0.set_samp_rate(samp_rate)
        if center_freq is not None:
            grblock.source_0.set_center_freq(center_freq, 0)
        grblock.source_0.set_gain(gain, 0)
        grblock.source_0.set_rx_agc(agc, 0)
        grblock.freq_setter = lambda x, y: x.set_center_freq(y, 0)
        grblock.set_command_time = lambda x, y: x.set_command_time(uhd.time_spec_t(y))
        grblock.clear_command_time = lambda x: x.clear_command_time()
        grblock.get_sdr_time_now = lambda x: x.get_time_now().get_real_secs()
        grblock.cmd_port = "command"
        return

    dev = f"driver={sdr}"
    stream_args = ""
    tune_args = [""]
    settings = [""]
    if sdrargs:
        settings = sdrargs
    grblock.source_0 = soapy.source(
        dev, "fc32", 1, "", stream_args, tune_args, settings
    )
    grblock.source_0.set_sample_rate(0, samp_rate)
    grblock.source_0.set_bandwidth(0, samp_rate)
    if center_freq is not None:
        grblock.source_0.set_frequency(0, center_freq)
    grblock.source_0.set_frequency_correction(0, 0)
    grblock.source_0.set_gain_mode(0, agc)
    grblock.source_0.set_gain(0, gain)
    grblock.freq_setter = lambda x, y: x.set_frequency(0, y)
    grblock.cmd_port = "cmd"
    return



"""Main entrypoint for GamutRF"""
from gamutrf.api import main as api_main
from gamutrf.freqxlator import main as freqxlator_main
from gamutrf.samples2raw import main as samples2raw_main
from gamutrf.scan import main as scan_main
from gamutrf.scan2mp4 import main as scan2mp4_main
from gamutrf.scan2rtlpow import main as scan2rtlpow_main
from gamutrf.sigfinder import main as sigfinder_main
from gamutrf.specgram import main as specgram_main


def api():
    """Entrypoint for API"""
    api_main()


def freqxlator():
    """Entrypoint for freqxlator"""
    freqxlator_main()


def samples2raw():
    """Entrypoint for samples2raw"""
    samples2raw_main()


def scan():
    """Entrypoint for scan"""
    scan_main()


def scan2mp4():
    """Entrypoint for scan2mp4"""
    scan2mp4_main()


def scan2rtlpow():
    """Entrypoint for scan2rtlpow"""
    scan2rtlpow_main()


def sigfinder():
    """Entrypoint for sigfinder"""
    sigfinder_main()


def specgram():
    """Entrypoint for specgram"""
    specgram_main()



import logging
import signal
import sys
import time
from argparse import ArgumentParser

try:
    from gnuradio import iqtlabs  # pytype: disable=import-error
    from gnuradio import eng_notation  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio.eng_arg import eng_float  # pytype: disable=import-error
    from gnuradio.eng_arg import intx  # pytype: disable=import-error
except ModuleNotFoundError:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme)"
    )
    sys.exit(1)

from prometheus_client import Gauge
from prometheus_client import start_http_server

from gamutrf.grscan import grscan


def init_prom_vars():
    prom_vars = {
        "freq_start_hz": Gauge("freq_start_hz", "start of scanning range in Hz"),
        "freq_end_hz": Gauge("freq_end_hz", "end of scanning range in Hz"),
        "sweep_sec": Gauge("sweep_sec", "scan sweep rate in seconds"),
    }
    return prom_vars


def argument_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "--freq-end",
        dest="freq_end",
        type=eng_float,
        default=eng_notation.num_to_str(float(1e9)),
        help="Set freq_end [default=%(default)r]",
    )
    parser.add_argument(
        "--freq-start",
        dest="freq_start",
        type=eng_float,
        default=eng_notation.num_to_str(float(100e6)),
        help="Set freq_start [default=%(default)r]",
    )
    parser.add_argument(
        "--igain",
        dest="igain",
        type=intx,
        default=0,
        help="Set igain[default=%(default)r]",
    )
    parser.add_argument(
        "--samp-rate",
        dest="samp_rate",
        type=eng_float,
        default=eng_notation.num_to_str(float(4.096e6)),
        help="Set samp_rate [default=%(default)r]",
    )
    parser.add_argument(
        "--sweep-sec",
        dest="sweep_sec",
        type=float,
        default=30,
        help="Set sweep_sec [default=%(default)r]",
    )
    parser.add_argument(
        "--tune-step-fft",
        dest="tune_step_fft",
        type=int,
        default=0,
        help="tune FFT step (0 is use sweep_sec) [default=%(default)r]",
    )
    parser.add_argument(
        "--skip-tune-step-fft",
        dest="skip_tune_step_fft",
        type=int,
        default=0,
        help="skip FFT samples on retune [default=%(default)r]",
    )
    parser.add_argument(
        "--nfft",
        dest="nfft",
        type=int,
        default=2048,
        help="FFTI size [default=%(default)r]",
    )
    parser.add_argument(
        "--logaddr",
        dest="logaddr",
        type=str,
        default="0.0.0.0",  # nosec
        help="Log FFT results to this address",
    )
    parser.add_argument(
        "--logport",
        dest="logport",
        type=int,
        default=8001,
        help="Log FFT results to this port",
    )
    parser.add_argument(
        "--promport",
        dest="promport",
        type=int,
        default=9000,
        help="Prometheus client port",
    )
    parser.add_argument(
        "--sdr",
        dest="sdr",
        type=str,
        default="ettus",
        help="SDR to use (ettus, bladerf, or lime)",
    )
    parser.add_argument(
        "--sdrargs",
        dest="sdrargs",
        type=str,
        default="",
        help="extra args to pass to SDR driver",
    )
    parser.add_argument(
        "--updatetimeout",
        dest="updatetimeout",
        type=int,
        default=10,
        help="seconds to wait for healthy freq updates",
    )
    parser.add_argument(
        "--tuneoverlap",
        dest="tuneoverlap",
        type=float,
        default=0.5,
        help="multiple of samp_rate when retuning",
    )
    return parser


def main():
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")
    options = argument_parser().parse_args()
    if gr.enable_realtime_scheduling() != gr.RT_OK:
        print("Warning: failed to enable real-time scheduling.")

    if options.freq_start > options.freq_end:
        print("Error: freq_start must be less than freq_end")
        sys.exit(1)

    if options.freq_end > 6e9:
        print("Error: freq_end must be less than 6GHz")
        sys.exit(1)

    if options.freq_start < 70e6:
        print("Error: freq_start must be at least 70MHz")
        sys.exit(1)

    # ensure tuning tags arrive on FFT window boundaries.
    if options.samp_rate % options.nfft:
        print("NFFT must be a factor of sample rate")
        sys.exit(1)

    prom_vars = init_prom_vars()
    prom_vars["freq_start_hz"].set(options.freq_start)
    prom_vars["freq_end_hz"].set(options.freq_end)
    prom_vars["sweep_sec"].set(options.sweep_sec)
    start_http_server(options.promport)

    tb = grscan(
        freq_end=options.freq_end,
        freq_start=options.freq_start,
        igain=options.igain,
        samp_rate=options.samp_rate,
        sweep_sec=options.sweep_sec,
        logaddr=options.logaddr,
        logport=options.logport,
        sdr=options.sdr,
        sdrargs=options.sdrargs,
        fft_size=options.nfft,
        tune_overlap=options.tuneoverlap,
        tune_step_fft=options.tune_step_fft,
        skip_tune_step_fft=options.skip_tune_step_fft,
        iqtlabs=iqtlabs,
    )

    def sig_handler(_sig=None, _frame=None):
        tb.stop()
        tb.wait()
        sys.exit(0)

    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    tb.start()
    tb.wait()
